import numpy as np
from scipy.integrate import odeint

class Century:

    ACTIVE_CONTAINER_INDEX = 3
    RESPIRATION_CONTAINER_INDEX = 5


    def __init__(self, Necromasa , LN , FraLA ,LENo ,PASo , ACTo , RESPo):
        self.Necromasa = Necromasa
        self.LN = LN
        self.FraLA = FraLA
        self.fs = 0.85 - (0.018 * self.LN)
        self.Ftex = 0.85 - (0.68 * self.FraLA)
        self.ESTo = self.Necromasa * (1 - self.fs)
        self.METo = self.Necromasa * self.fs
        self.LENo = LENo
        self.PASo = PASo
        self.ACTo = ACTo
        self.RESPo = RESPo


    def resolve(self, time, Kmet, Kest , Kminl, Khumac , Kminp , ResEL , ResEA , ResMet , ResLA , ResPA , PartEst , PartLen , PartAct ):

        y0 = self.ESTo, self.METo, self.LENo, self.ACTo, self.PASo, self.RESPo
        
        args = (self.Ftex, Kmet, Kest, Kminl, Khumac, Kminp, ResEL, ResEA, ResMet, ResLA, ResPA, PartEst, PartLen, PartAct)
        
        y = odeint(Century.ode, y0, time, args)
        
        return ([row[Century.ACTIVE_CONTAINER_INDEX] for row in y ], [row[Century.RESPIRATION_CONTAINER_INDEX] for row in y ])
    

    @staticmethod
    def ode(y, time, Ftex, Kmet, Kest, Kminl, Khumac, Kminp, ResEL, ResEA, ResMet, ResLA, ResPA, PartEst, PartLen, PartAct):
        ESTo, METo, LENTo, ACTo, PASo, RESPo = y
        
        dESTdt = -ESTo * Kest 
        
        dMETdt = -METo * Kmet 
        
        dLENTdt = ((1 - ResEL) * PartEst * ESTo * Kest)  \
                + ((1 - Ftex - PartAct) * ACTo * Khumac) \
                - (LENTo * Kminl)
        
        dACTdt =  ((1 - ResEA) * (1 -PartEst) * ESTo * Kest) \
                + ((1 - ResMet) * METo * Kmet) \
                + ((1 - ResLA - PartLen) * LENTo * Kminl) \
                + ((1 - ResPA) * PASo * Kminp) \
                - (ACTo * Khumac)
        
        dPASdt =  (PartAct * ACTo * Khumac) \
                + (PartLen * LENTo * Kminl) \
                - (PASo * Kminp) 
        
        dRESPdt = (ResEL * PartEst * ESTo * Kest) \
                + (ResEA * (1 - PartEst) * ESTo * Kest) \
                + (Ftex * ACTo * Khumac) \
                + (ResMet * METo * Kmet) \
                + (ResLA * LENTo * Kminl) \
                + (ResPA * PASo * Kminp)
        
        return (dESTdt, dMETdt, dLENTdt, dACTdt, dPASdt, dRESPdt)



import numpy as np
import time
import sys
import os

class Genetic:
    
    
    def __init__(self, century_model, time_data, training_data):
        self.century_model = century_model
        self.time_data = time_data
        self.training_data = training_data
        self.population_size = 200
        self.selection_rate = 0.3
        self.mutation_rate  = 0.01
        self.gen_len = 3
        self.n_params = 13
        self.generation = 0
        self.error = sys.maxsize
        self.start_time = time.time()
        self.population = []
        self.milestones = []
        self.fitness = []
        self.generate_population()


    def mse(self, parameters):
        model_data = self.century_model.resolve(self.time_data,
                                                parameters[0], 
                                                parameters[1], 
                                                parameters[2], 
                                                parameters[3], 
                                                parameters[4], 
                                                parameters[5], 
                                                parameters[6], 
                                                parameters[7], 
                                                parameters[8], 
                                                parameters[9], 
                                                parameters[10], 
                                                parameters[11], 
                                                parameters[12])
        max_value = 0
        for i in [0,1]:
            values_model = model_data[i]
            values_training = self.training_data[i]
            sum = 0
            for j in range(len(values_model)):
                sum += (values_model[j] - values_training[j]) ** 2
            avg = (sum / len(values_model))
            if(avg > max_value):
                max_value = avg
        return round(max_value)

    def generate_population(self):
        for i in range(self.population_size):
            child = list(np.random.rand(self.n_params))
            self.population.append(child)

    def mutate(self):
        for i in range(len(self.population)):
            child = self.population[i]
            if np.random.random() < self.mutation_rate:
                gen = list(np.random.rand(self.gen_len))
                index = np.random.randint(len(child) - self.gen_len + 1)
                child = child[0:index] + gen + child[index + self.gen_len:]
            self.population[i] = child

    def evaluate_population(self):
        fitness_tmp = []
        for child in self.population:
            y = self.mse(child)
            fitness_tmp += [y]
        self.fitness = np.array(fitness_tmp)

    def update_error(self):
        fitness_sorted = sorted(self.fitness)
        error = fitness_sorted[0]
        if(error < self.error and error >= 0):
            self.error = error
            self.milestones.append([self.generation, self.error])
            return True
        return False

    def choose_parents(self):
        fitness_tmp = []
        population_tmp = []
        for i in range(int(self.population_size * self.selection_rate)):
            index = np.random.choice(self.population_size)
            fitness_tmp.append(self.fitness[index])
            population_tmp.append(self.population[index])
        parents = [x for _,x in sorted(zip(fitness_tmp,population_tmp))]
        return (parents[0],parents[1])

    def select_offspring(self):
        offspring = []
        for i in range(self.population_size//2):
            parents =  self.choose_parents()
            cross_point = np.random.randint(self.n_params)
            offspring += [ parents[0][:cross_point] + parents[1][cross_point:] ]
            offspring += [ parents[1][:cross_point] + parents[0][cross_point:] ]
        return offspring


    def evolve(self, generations):
        for _ in range(generations):
            self.mutate()
            self.evaluate_population()
            if self.update_error():
                self.population = self.select_offspring()
            self.print_status()
            self.generation += 1
        self.print_status(True)

    def format_time(self, seconds):
        hh = int(seconds // 3600)
        mm = int((seconds - (hh * 3600)) // 60)
        ss = int((seconds - (hh * 3600) - (mm * 60)))
        return f'{ "0" if hh < 10 else "" }{hh}:{"0" if mm < 10 else ""}{mm}:{"0" if ss < 10 else ""}{ss}'

    def print_status(self,done = False):
        time_formatted = self.format_time(time.time() - self.start_time)
        error_str = '{}'.format(self.error).rjust(10," ")
        if done: 
            print(f'DONE! => time: {time_formatted}, generation: {self.generation}, error: {error_str}')     
        else:
            print(f'time: {time_formatted}, generation: {self.generation}, error: {error_str}', end='\r') 


    def solution(self):
        index = -1
        min_value = sys.maxsize
        for i in range(len(self.population)):
            child = self.population[i]
            value = self.mse(child)
            if value < min_value:
                min_value = value
                index = i
        return self.population[index]



import pandas as pd
import matplotlib.pyplot as plt
import os.path
from centurypy.century import Century
from centurypy.genetic import Genetic

class  CenturyPy:

    def __init__(self, input_file,  input_values):
        if not self.__check_input_file(input_file):
           raise Exception("Input file not found")
        if not self.__check_input_values(input_values):
           raise Exception("Wrong input values format")
        self.params_name = ['Kmet', 'Kest', 'Kminl', 'Khumac', 'Kminp', 'ResEL', 
                            'ResEA', 'ResMet', 'ResLA', 'ResPA', 'PartEst', 'PartLen','PartAct']
        self.training_data = []
        self.time_data = []
        self.century_model = None
        self.genetic_model = None
        self.__load_input_file(input_file)
        self.__load_century_model(input_values)
        self.__load_genetic_model()
        
    def __check_input_file(self, input_file):
        return os.path.isfile(input_file)

    def __check_input_values(self, input):
        if not type(input) is dict: 
            return False
        if input.get('necromasa') == None:
            return False
        if input.get('ln') == None:
            return False
        if input.get('frala') == None:
            return False
        if input.get('leno') == None:
            return False
        if input.get('paso') == None:
            return False
        if input.get('acto') == None:
            return False
        if input.get('respo') == None:
            return False
        return True

    def __load_input_file(self, input_file):
        df = pd.read_csv(input_file,header=None)
        array = df.to_numpy()
        self.training_data = (array[:,1][:],array[:,2][:])
        self.time_data = array[:,0][:]

    def __load_century_model(self,input_values):
        self.century_model = Century(
            Necromasa = input_values['necromasa'], 
            ACTo=input_values['acto'], 
            LENo=input_values['leno'], 
            PASo=input_values['paso'], 
            RESPo= input_values['respo'] , 
            LN= input_values['ln'], 
            FraLA= input_values['frala']
        )

    def __load_genetic_model(self):
        self.genetic_model = Genetic(self.century_model,
                                     self.time_data,
                                     self.training_data)

    def __save_parameters(self, parameters):
        f = open("parameters.txt","w")
        for i in range(len(parameters)):
            f.write(f'{self.params_name[i]}: {parameters[i]}\n')
        f.close()

    def __display_parameters(self, parameters):
        print('+++++++++++++++++++++++++++++++++++++++++++++++')
        print('+++++++++++++++ PARAMETERS ++++++++++++++++++++')
        for i in range(len(parameters)):
            param_name = self.params_name[i]
            print(f'{param_name.rjust(10," ")}: {parameters[i]}')
        print('+++++++++++++++++++++++++++++++++++++++++++++++')

    def __display_chart(self,parameters):
        generations, error = zip(*self.genetic_model.milestones)
        solution = self.century_model.resolve(self.time_data,
                                              parameters[0], 
                                              parameters[1], 
                                              parameters[2], 
                                              parameters[3], 
                                              parameters[4], 
                                              parameters[5], 
                                              parameters[6], 
                                              parameters[7], 
                                              parameters[8], 
                                              parameters[9], 
                                              parameters[10], 
                                              parameters[11], 
                                              parameters[12])
        fig, axes = plt.subplots(2,1)
        colors = ['b','r']
        names = ['Active','Respiration']
        for i in [0,1]:
            name = names[i]
            color = colors[i]
            axes[0].plot(self.time_data,
                         self.training_data[i],
                         f'{color}.' , 
                         label = f'{name} (Original)')
            axes[0].plot(self.time_data,
                         solution[i],
                         f'{color}-' , 
                         label = f'{name} (Solution)')
        axes[0].set(xlabel='Time',ylabel='C(mg)')
        axes[0].legend()
        axes[1].plot(generations,error,'b.-')
        axes[1].set(xlabel='Generations',ylabel='Error (MSE)')
        plt.show()

    def fit_model(self, generations = 500, chart = True):
        self.genetic_model.evolve(generations)
        params = self.genetic_model.solution()
        self.__save_parameters(params)
        self.__display_parameters(params)
        if chart:
            self.__display_chart(params)

